@@ -98,7 +98,7 @@ def get_dataset_size(dataset, batch_size=5000):
   Returns:
     The dataset size or ``None`` if the dataset is infinite.
   """
-  if tf.data.experimental.cardinality(dataset) == tf.data.experimental.INFINITE_CARDINALITY:
+  if dataset.cardinality() == tf.data.INFINITE_CARDINALITY:
     return None
   if batch_size is not None:
     dataset = dataset.batch(batch_size)
@@ -225,7 +225,7 @@ def make_cardinality_multiple_of(divisor):
 
   def _transform(dataset):
     # Nothing to do for infinite datasets.
-    if tf.data.experimental.cardinality(dataset) == tf.data.experimental.INFINITE_CARDINALITY:
+    if dataset.cardinality() == tf.data.INFINITE_CARDINALITY:
       return dataset
 
     # Concatenate extra batches with a flag.
@@ -377,6 +377,8 @@ def batch_sequence_dataset(batch_size,
 
   Raises:
     ValueError: if :obj:`batch_type` is not one of "examples" or "tokens".
+    ValueError: if :obj:`batch_type` is "tokens" but :obj:`length_bucket_width`
+      is not set.
     ValueError: if the number of length functions in :obj:`length_fn` does not
       match the number of parallel elements.
 
@@ -425,6 +427,9 @@ def batch_sequence_dataset(batch_size,
     return tf.cast(tf.maximum(size, required_multiple), tf.int64)
 
   if length_bucket_width is None:
+    if batch_type == "tokens":
+      raise ValueError("Batch type 'tokens' requires length bucketing (the parameter "
+                       "length_bucket_width should be non null)")
     return batch_dataset(batch_size, padded_shapes=padded_shapes)
 
   if batch_type == "examples":
@@ -474,7 +479,7 @@ def training_pipeline(batch_size,
 
   Args:
     batch_size: The batch size to use.
-    batch_type: The training batching stragety to use: can be "examples" or
+    batch_type: The training batching strategy to use: can be "examples" or
       "tokens".
     batch_multiplier: The batch size multiplier.
     batch_size_multiple: When :obj:`batch_type` is "tokens", ensure that the
@@ -585,6 +590,7 @@ def training_pipeline(batch_size,
   return _pipeline
 
 def inference_pipeline(batch_size,
+                       batch_type="examples",
                        process_fn=None,
                        transform_fns=None,
                        length_bucket_width=None,
@@ -599,6 +605,7 @@ def inference_pipeline(batch_size,
 
   Args:
     batch_size: The batch size to use.
+    batch_type: The batching strategy to use: can be "examples" or "tokens".
     process_fn: The processing function to apply on each element.
     transform_fns: List of dataset transformation functions (applied after
       :obj:`process_fn` if defined).
@@ -622,7 +629,11 @@ def inference_pipeline(batch_size,
   """
 
   def _inject_index(index, x):
-    x["index"] = index
+    if isinstance(x, tuple):
+      features = x[0]
+    else:
+      features = x
+    features["index"] = index
     return x
 
   def _pipeline(dataset):
@@ -634,12 +645,20 @@ def inference_pipeline(batch_size,
     if length_bucket_width is not None and length_bucket_width > 0:
       if length_fn is None:
         raise ValueError("length_fn is required when reordering by length")
-      if not isinstance(_get_output_shapes(dataset), dict):
+      output_shapes = _get_output_shapes(dataset)
+      if isinstance(output_shapes, tuple):
+        num_length_fn = len(length_fn) if isinstance(length_fn, (list, tuple)) else 1
+        if len(output_shapes) != num_length_fn:
+          raise ValueError("The dataset outputs %d parallel features, but got %d "
+                           "length functions" % (len(output_shapes), num_length_fn))
+        output_shapes = output_shapes[0]
+      if not isinstance(output_shapes, dict):
         raise ValueError("Reordering by length expects dataset elements to be Python dicts")
       dataset = dataset.enumerate()
       dataset = dataset.map(_inject_index)
       dataset = dataset.apply(batch_sequence_dataset(
           batch_size,
+          batch_type=batch_type,
           length_bucket_width=length_bucket_width,
           length_fn=length_fn))
     else:
@@ -648,39 +667,3 @@ def inference_pipeline(batch_size,
     return dataset
 
   return _pipeline
-
-def function_on_next(dataset, as_numpy=False):
-  """Decorator to run a ``tf.function`` on each dataset element.
-
-  The motivation for this construct is to get the next element within the
-  ``tf.function`` for increased efficiency.
-
-  Args:
-    dataset: The dataset to iterate.
-    as_numpy: If ``True``, call `.numpy()` on each output tensors.
-
-  Returns:
-    A function decorator. The decorated function is transformed into a callable
-    that returns a generator over its outputs.
-  """
-
-  def decorator(func):
-    def _fun():
-      iterator = iter(dataset)
-
-      @tf.function
-      def _tf_fun():
-        return func(lambda: next(iterator))
-
-      while True:
-        try:
-          outputs = _tf_fun()
-          if as_numpy:
-            outputs = tf.nest.map_structure(lambda x: x.numpy(), outputs)
-          yield outputs
-        except tf.errors.OutOfRangeError:
-          break
-
-    return _fun
-
-  return decorator
