@@ -28,6 +28,8 @@ class Evaluator(object):
                features_file,
                labels_file,
                batch_size,
+               batch_type="examples",
+               length_bucket_width=None,
                scorers=None,
                save_predictions=False,
                early_stopping=None,
@@ -38,10 +40,14 @@ class Evaluator(object):
     """Initializes the evaluator.
 
     Args:
-      model: A :class:`opennmt.models.model.Model` to evaluate.
+      model: A :class:`opennmt.models.Model` to evaluate.
       features_file: Path to the evaluation features.
       labels_file: Path to the evaluation labels.
       batch_size: The evaluation batch size.
+      batch_type: The batching strategy to use: can be "examples" or "tokens".
+      length_bucket_width: The width of the length buckets to select batch
+        candidates from (for efficiency). Set ``None`` to not constrain batch
+        formation.
       scorers: A list of scorers, callables taking the path to the reference and
         the hypothesis and return one or more scores.
       save_predictions: Save evaluation predictions to a file. This is ``True``
@@ -99,6 +105,8 @@ class Evaluator(object):
         features_file,
         labels_file,
         batch_size,
+        batch_type=batch_type,
+        length_bucket_width=length_bucket_width,
         num_threads=1,
         prefetch_buffer_size=1)
 
@@ -130,7 +138,7 @@ class Evaluator(object):
     """Creates an evaluator from the configuration.
 
     Args:
-      model: A :class:`opennmt.models.model.Model` to evaluate.
+      model: A :class:`opennmt.models.Model` to evaluate.
       config: The global user configuration.
       features_file: Optional input features file to evaluate. If not set, will
         load ``eval_features_file`` from the data configuration.
@@ -141,10 +149,14 @@ class Evaluator(object):
       A :class:`opennmt.evaluation.Evaluator` instance.
 
     Raises:
-      ValueError: if one of :obj:`features_file` and :obj:`labels_file` is set
-        but not the other.
+      ValueError: for supervised models, if one of :obj:`features_file` and
+        :obj:`labels_file` is set but not the other.
+      ValueError: for unsupervised models, if :obj:`labels_file` is set.
     """
-    if (features_file is None) != (labels_file is None):
+    if model.unsupervised:
+      if labels_file is not None:
+        raise ValueError("labels_file can not be set when evaluating unsupervised models")
+    elif (features_file is None) != (labels_file is None):
       raise ValueError("features_file and labels_file should be both set for evaluation")
     eval_config = config["eval"]
     scorers = eval_config.get("external_evaluators")
@@ -163,6 +175,8 @@ class Evaluator(object):
         features_file or config["data"]["eval_features_file"],
         labels_file or config["data"].get("eval_labels_file"),
         eval_config["batch_size"],
+        batch_type=eval_config.get("batch_type", "examples"),
+        length_bucket_width=eval_config.get("length_bucket_width"),
         scorers=scorers,
         save_predictions=eval_config.get("save_eval_predictions", False),
         early_stopping=early_stopping,
@@ -266,6 +280,10 @@ class Evaluator(object):
     if self._save_predictions:
       output_path = os.path.join(self._eval_dir, "predictions.txt.%d" % step)
       output_file = tf.io.gfile.GFile(output_path, "w")
+      write_fn = lambda prediction: (
+          self._model.print_prediction(prediction, stream=output_file))
+      index_fn = lambda prediction: prediction.get("index")
+      ordered_writer = misc.OrderRestorer(index_fn, write_fn)
 
     loss_num = 0
     loss_den = 0
@@ -283,7 +301,7 @@ class Evaluator(object):
       if output_file is not None:
         predictions = {k:v.numpy() for k, v in predictions.items()}
         for prediction in misc.extract_batches(predictions):
-          self._model.print_prediction(prediction, stream=output_file)
+          ordered_writer.push(prediction)
     if loss_den == 0:
       raise RuntimeError("No examples were evaluated")
     loss = loss_num / loss_den
