@@ -81,6 +81,7 @@ class Inputter(tf.keras.layers.Layer):
   def make_inference_dataset(self,
                              features_file,
                              batch_size,
+                             batch_type="examples",
                              length_bucket_width=None,
                              num_threads=1,
                              prefetch_buffer_size=None):
@@ -92,6 +93,7 @@ class Inputter(tf.keras.layers.Layer):
     Args:
       features_file: The test file.
       batch_size: The batch size to use.
+      batch_type: The batching strategy to use: can be "examples" or "tokens".
       length_bucket_width: The width of the length buckets to select batch
         candidates from (for efficiency). Set ``None`` to not constrain batch
         formation.
@@ -118,6 +120,7 @@ class Inputter(tf.keras.layers.Layer):
     dataset = self.make_dataset(features_file, training=False)
     dataset = dataset.apply(dataset_util.inference_pipeline(
         batch_size,
+        batch_type=batch_type,
         transform_fns=transform_fns,
         length_bucket_width=length_bucket_width,
         length_fn=self.get_length,
@@ -314,8 +317,26 @@ class ParallelInputter(MultiInputter):
         if type(inputter) is not type(leaves[0]):
           raise ValueError("Each inputter must be of the same type for parameter sharing")
 
+  def _structure(self):
+    """Returns the nested structure that represents this parallel inputter."""
+    return [
+        inputter._structure()  # pylint: disable=protected-access
+        if isinstance(inputter, ParallelInputter) else None
+        for inputter in self.inputters]
+
   def make_dataset(self, data_file, training=None):
-    if not isinstance(data_file, list) or len(data_file) != len(self.inputters):
+    if not isinstance(data_file, list):
+      data_file = [data_file]
+
+    # For evaluation and inference, accept a flat list of data files for nested inputters.
+    # This is needed when nesting can't easily be represented (e.g. on the command line).
+    if not training:
+      try:
+        data_file = tf.nest.pack_sequence_as(self._structure(), tf.nest.flatten(data_file))
+      except ValueError:
+        data_file = []  # This will raise the error below.
+
+    if len(data_file) != len(self.inputters):
       raise ValueError("The number of data files must be the same as the number of inputters")
 
     num_files = -1
@@ -506,6 +527,8 @@ class ExampleInputterAdapter:
                               features_file,
                               labels_file,
                               batch_size,
+                              batch_type="examples",
+                              length_bucket_width=None,
                               num_threads=1,
                               prefetch_buffer_size=None):
     """Builds a dataset to be used for evaluation.
@@ -514,6 +537,10 @@ class ExampleInputterAdapter:
       features_file: The evaluation source file.
       labels_file: The evaluation target file.
       batch_size: The batch size to use.
+      batch_type: The batching strategy to use: can be "examples" or "tokens".
+      length_bucket_width: The width of the length buckets to select batch
+        candidates from (for efficiency). Set ``None`` to not constrain batch
+        formation.
       num_threads: The number of elements processed in parallel.
       prefetch_buffer_size: The number of batches to prefetch asynchronously. If
         ``None``, use an automatically tuned value.
@@ -526,8 +553,10 @@ class ExampleInputterAdapter:
     """
     if labels_file is not None:
       data_files = [features_file, labels_file]
+      length_fn = [self.features_inputter.get_length, self.labels_inputter.get_length]
     else:
       data_files = features_file
+      length_fn = self.get_length
 
     map_fn = lambda *arg: self.make_features(element=misc.item_or_tuple(arg), training=False)
     transform_fns = [lambda dataset: dataset.map(map_fn, num_parallel_calls=num_threads or 1)]
@@ -535,7 +564,10 @@ class ExampleInputterAdapter:
     dataset = self.make_dataset(data_files, training=False)
     dataset = dataset.apply(dataset_util.inference_pipeline(
         batch_size,
+        batch_type=batch_type,
         transform_fns=transform_fns,
+        length_bucket_width=length_bucket_width,
+        length_fn=length_fn,
         num_threads=num_threads,
         prefetch_buffer_size=prefetch_buffer_size))
     return dataset
@@ -571,7 +603,7 @@ class ExampleInputterAdapter:
       features_file: The source file or a list of training source files.
       labels_file: The target file or a list of training target files.
       batch_size: The batch size to use.
-      batch_type: The training batching stragety to use: can be "examples" or
+      batch_type: The training batching strategy to use: can be "examples" or
         "tokens".
       batch_multiplier: The batch size multiplier to prepare splitting accross
          replicated graph parts.
@@ -667,12 +699,14 @@ class ExampleInputter(ParallelInputter, ExampleInputterAdapter):
   def make_inference_dataset(self,
                              features_file,
                              batch_size,
+                             batch_type="examples",
                              length_bucket_width=None,
                              num_threads=1,
                              prefetch_buffer_size=None):
     return self.features_inputter.make_inference_dataset(
         features_file,
         batch_size,
+        batch_type=batch_type,
         length_bucket_width=length_bucket_width,
         num_threads=num_threads,
         prefetch_buffer_size=prefetch_buffer_size)
