@@ -11,15 +11,17 @@ from opennmt.layers.reducer import SumReducer
 class PositionEncoder(tf.keras.layers.Layer):
   """Base class for position encoders."""
 
-  def __init__(self, reducer=SumReducer(), **kwargs):
+  def __init__(self, reducer=None, **kwargs):
     """Initializes the position encoder.
 
     Args:
       reducer: A :class:`opennmt.layers.Reducer` to merge inputs and position
-        encodings.
+        encodings. Defaults to :class:`opennmt.layers.SumReducer`.
       **kwargs: Additional layer keyword arguments.
     """
     super(PositionEncoder, self).__init__(**kwargs)
+    if reducer is None:
+      reducer = SumReducer(dtype=kwargs.get("dtype"))
     self.reducer = reducer
 
   def call(self, inputs, position=None):  # pylint: disable=arguments-differ
@@ -58,14 +60,14 @@ class PositionEncoder(tf.keras.layers.Layer):
 class PositionEmbedder(PositionEncoder):
   """Encodes position with a lookup table."""
 
-  def __init__(self, maximum_position=128, reducer=SumReducer(), **kwargs):
+  def __init__(self, maximum_position=128, reducer=None, **kwargs):
     """Initializes the position encoder.
 
     Args:
       maximum_position: The maximum position to embed. Positions greater
         than this value will be set to :obj:`maximum_position`.
       reducer: A :class:`opennmt.layers.Reducer` to merge inputs and position
-        encodings.
+        encodings. Defaults to :class:`opennmt.layers.SumReducer`.
       **kwargs: Additional layer keyword arguments.
     """
     super(PositionEmbedder, self).__init__(reducer=reducer, **kwargs)
@@ -97,7 +99,7 @@ class SinusoidalPositionEncoder(PositionEncoder):
 
     log_timescale_increment = math.log(10000) / (depth / 2 - 1)
     inv_timescales = tf.exp(tf.range(depth / 2, dtype=tf.float32) * -log_timescale_increment)
-    inv_timescales = tf.reshape(tf.tile(inv_timescales, [batch_size]), [batch_size, -1])
+    inv_timescales = tf.reshape(tf.tile(inv_timescales, [batch_size]), [batch_size, depth // 2])
     scaled_time = tf.expand_dims(positions, -1) * tf.expand_dims(inv_timescales, 1)
     encoding = tf.concat([tf.sin(scaled_time), tf.cos(scaled_time)], axis=2)
     return tf.cast(encoding, self.dtype)
