@@ -14,6 +14,7 @@ import numpy as np
 import tensorflow as tf
 
 from opennmt import evaluation
+from opennmt import inference
 from opennmt import models
 from opennmt import training as training_util
 from opennmt.utils import checkpoint as checkpoint_util
@@ -30,10 +31,13 @@ _CONFIG_FALLBACK = {
         "save_summary_steps": 100
     },
     "eval": {
+        "length_bucket_width": None,
+        "batch_type": "examples",
         "batch_size": 32
     },
     "infer": {
         "length_bucket_width": None,
+        "batch_type": "examples",
         "batch_size": 16
     },
     "score": {
@@ -53,14 +57,27 @@ class Runner(object):
     """Initializes the runner parameters.
 
     Args:
-      model: A :class:`opennmt.models.Model` instance to run.
+      model: A :class:`opennmt.models.Model` instance to run or a callable that
+        returns such instance.
       config: The run configuration.
       auto_config: If ``True``, use automatic configuration values defined by
         :obj:`model`.
       mixed_precision: Enable mixed precision.
       seed: The random seed to set.
+
+    Raises:
+      TypeError: if :obj:`model` is not a :class:`opennmt.models.Model` instance
+        or a callable.
     """
-    self._model = model
+    if isinstance(model, models.Model):
+      self._model = model
+      self._model_fn = lambda: misc.clone_layer(model)
+    elif callable(model):
+      self._model = model()
+      self._model_fn = model
+    else:
+      raise TypeError("model should be a opennmt.models.Model instance or a callable")
+    tf.get_logger().info("Using model:\n%s", self._model)
     self._optimizer = None
     self._config = copy.deepcopy(config)
     self._auto_config = auto_config
@@ -117,7 +134,7 @@ class Runner(object):
     return config
 
   def _init_model(self, config):
-    model = misc.clone_layer(self._model)
+    model = self._model_fn()
     model.initialize(config["data"], params=config["params"])
     return model
 
@@ -201,10 +218,12 @@ class Runner(object):
         raise ValueError("num_devices (or num_gpus) should be set to 1 when using Horovod")
       trainer = training_util.HorovodTrainer(
           model, optimizer, hvd, checkpoint=checkpoint)
-    else:
+    elif num_devices > 1:
       devices = misc.get_devices(count=num_devices)
-      trainer = training_util.DistributionStrategyTrainer(
+      trainer = training_util.MirroredStrategyTrainer(
           model, optimizer, checkpoint=checkpoint, devices=devices)
+    else:
+      trainer = training_util.Trainer(model, optimizer, checkpoint=checkpoint)
 
     trainer(
         dataset_fn,
@@ -242,7 +261,7 @@ class Runner(object):
     model = self._init_model(config)
     checkpoint = checkpoint_util.Checkpoint.from_config(config, model)
     checkpoint_path = checkpoint.restore(checkpoint_path=checkpoint_path, weights_only=True)
-    step = int(checkpoint_path.split("-")[-1])
+    step = checkpoint_util.get_step_from_checkpoint_prefix(checkpoint_path)
     evaluator = evaluation.Evaluator.from_config(
         model,
         config,
@@ -324,7 +343,8 @@ class Runner(object):
 
     Args:
       features_file: The file(s) to infer from.
-      predictions_file: If set, predictions are saved in this file.
+      predictions_file: If set, predictions are saved in this file, otherwise
+        they are printed on the standard output.
       checkpoint_path: Path of a specific checkpoint to predict. If ``None``,
         the latest is used.
       log_time: If ``True``, several time metrics will be printed in the logs at
@@ -338,74 +358,15 @@ class Runner(object):
     dataset = model.examples_inputter.make_inference_dataset(
         features_file,
         infer_config["batch_size"],
+        batch_type=infer_config["batch_type"],
         length_bucket_width=infer_config["length_bucket_width"],
         prefetch_buffer_size=infer_config.get("prefetch_buffer_size"))
-
-    if predictions_file:
-      stream = open(predictions_file, encoding="utf-8", mode="w")
-    else:
-      stream = sys.stdout
-
-    infer_fn = tf.function(model.infer, input_signature=(dataset.element_spec,))
-    tf.get_logger().info("Tracing and optimizing the inference graph...")
-    infer_fn.get_concrete_function()  # Trace the function now.
-
-    # Inference might return out-of-order predictions. The OrderRestorer utility is
-    # used to write predictions in their original order.
-    write_fn = lambda prediction: (
-        model.print_prediction(prediction, params=infer_config, stream=stream))
-    index_fn = lambda prediction: prediction.get("index")
-    ordered_writer = misc.OrderRestorer(index_fn, write_fn)
-
-    total_time = 0
-    total_tokens = 0
-    total_examples = 0
-    start_time = time.time()
-
-    # When the inference dataset is bucketized, it can happen that no output is
-    # written in a long time. To avoid confusion and give the impression that
-    # the process is stuck, we ensure that something is logged regularly.
-    max_time_without_output = 10
-    last_output_time = start_time
-
-    for source in dataset:
-      predictions = infer_fn(source)
-      predictions = tf.nest.map_structure(lambda t: t.numpy(), predictions)
-      batch_time = time.time()
-
-      for prediction in misc.extract_batches(predictions):
-        written = ordered_writer.push(prediction)
-        if written:
-          last_output_time = batch_time
-        else:
-          time_without_output = batch_time - last_output_time
-          if time_without_output >= max_time_without_output:
-            tf.get_logger().info(
-                "%d predictions are buffered, but waiting for the prediction of "
-                "line %d to advance the output...",
-                ordered_writer.buffer_size,
-                ordered_writer.next_index + 1)
-            last_output_time = batch_time
-
-      if log_time:
-        batch_size = next(iter(predictions.values())).shape[0]
-        total_examples += batch_size
-        length = predictions.get("length")
-        if length is not None:
-          if len(length.shape) == 2:
-            length = length[:, 0]
-          total_tokens += sum(length)
-
-    if log_time:
-      end_time = time.time()
-      total_time = end_time - start_time
-      tf.get_logger().info("Total prediction time (s): %f", total_time)
-      tf.get_logger().info(
-          "Average prediction time (s): %f", total_time / total_examples)
-      if total_tokens > 0:
-        tf.get_logger().info("Tokens per second: %f", total_tokens / total_time)
-    if predictions_file:
-      stream.close()
+    inference.predict_dataset(
+        model,
+        dataset,
+        print_params=infer_config,
+        predictions_file=predictions_file,
+        log_time=log_time)
 
   def export(self, export_dir, checkpoint_path=None, exporter=None):
     """Exports a model.
@@ -443,21 +404,7 @@ class Runner(object):
         predictions_file,
         score_config["batch_size"],
         prefetch_buffer_size=score_config.get("prefetch_buffer_size"))
-
-    if output_file:
-      stream = open(output_file, encoding="utf-8", mode="w")
-    else:
-      stream = sys.stdout
-
-    score_fn = tf.function(model.score, input_signature=dataset.element_spec)
-    for features, labels in dataset:
-      results = score_fn(features, labels)
-      results = tf.nest.map_structure(lambda t: t.numpy(), results)
-      for batch in misc.extract_batches(results):
-        model.print_score(batch, params=score_config, stream=stream)
-
-    if output_file:
-      stream.close()
+    inference.score_dataset(model, dataset, print_params=score_config, output_file=output_file)
 
 
 def _count_batch_accum(batch_size, target_batch_size, num_replicas=1):
