@@ -1,7 +1,7 @@
 """Training related classes and functions."""
 
-import abc
 import contextlib
+import itertools
 import time
 
 import tensorflow as tf
@@ -10,14 +10,17 @@ from opennmt.optimizers import utils as optimizer_util
 from opennmt.utils import misc
 
 
-class Trainer(abc.ABC):
-  """Base class for model trainer."""
+class Trainer:
+  """Base class for model trainer, implementing single-GPU training."""
 
   def __init__(self, model, optimizer, checkpoint=None, is_master=True):
     """Initializes the trainer.
 
     Args:
-      checkpoint: A :class:`opennmt.utils.Checkpoint` instance.
+      model: A :class:`opennmt.models.Model` instance to train.
+      optimizer: A ``tf.keras.optimizers.Optimizer`` instance.
+      checkpoint: A :class:`opennmt.utils.checkpoint.Checkpoint` instance. If
+        not set, no checkpoints will be saved.
       is_master: Whether this trainer instance is the master trainer.
     """
     self._checkpoint = checkpoint
@@ -36,7 +39,7 @@ class Trainer(abc.ABC):
     mixed_precision_enabled = graph_optimizer_options.get("auto_mixed_precision")
     if (mixed_precision_enabled
         and not isinstance(optimizer, tf.keras.mixed_precision.experimental.LossScaleOptimizer)):
-      optimizer = _LossScaleOptimizer(optimizer, "dynamic")
+      optimizer = tf.keras.mixed_precision.experimental.LossScaleOptimizer(optimizer, "dynamic")
     self._optimizer = optimizer
 
   @property
@@ -146,7 +149,20 @@ class Trainer(abc.ABC):
       evaluator(step)
       return evaluator.should_stop()
 
-  @abc.abstractmethod
+  def _finalize_dataset(self, dataset):
+    """Returns the final dataset instance to be used for training.
+
+    Args:
+      dataset: A ``tf.data.Dataset`` or a function taking a ``tf.distribute.InputContext``
+        instance and returning a ``tf.data.Dataset``.
+
+    Returns:
+      A ``tf.data.Dataset``.
+    """
+    if callable(dataset):
+      dataset = dataset(tf.distribute.InputContext())
+    return dataset
+
   def _steps(self, dataset, accum_steps=1, report_steps=None):
     """Returns a generator over training steps (i.e. parameters update).
 
@@ -159,7 +175,55 @@ class Trainer(abc.ABC):
     Returns:
       A generator that yields a loss value to report for this step.
     """
-    raise NotImplementedError()
+    dataset = self._finalize_dataset(dataset)
+    iterator = iter(dataset)
+
+    # Wrap forward and step with tf.function.
+    # We get the next dataset element within the function for increased efficiency.
+
+    if accum_steps is None or accum_steps == 1:
+
+      @tf.function
+      def _forward_and_step():
+        source, target = next(iterator)
+        return self._forward_and_step(
+            source,
+            target,
+            record_summaries=self._should_record_summaries(report_steps, with_accum=False))
+
+      for i in itertools.count():
+        try:
+          loss = _forward_and_step()
+        except tf.errors.OutOfRangeError:  # Dataset iterator exhausted.
+          break
+        if i == 0:
+          self._broadcast_variables()
+        yield loss
+
+    else:
+
+      @tf.function
+      def _forward():
+        source, target = next(iterator)
+        return self._forward(
+            source,
+            target,
+            record_summaries=self._should_record_summaries(report_steps, with_accum=True))
+
+      @tf.function
+      def _step():
+        return self._step()
+
+      for i in itertools.count():
+        try:
+          loss = _forward()
+        except tf.errors.OutOfRangeError:  # Dataset iterator exhausted.
+          break
+        if i == 0 or (i + 1) % accum_steps == 0:
+          _step()
+          if i == 0:
+            self._broadcast_variables()
+          yield loss
 
   def _run_model(self, source, target):
     """Computes the loss of the given source and target pair.
@@ -202,31 +266,56 @@ class Trainer(abc.ABC):
           len(self._model.non_trainable_weights))
     return training_loss, reported_loss
 
-  def _should_record_summaries(self, report_steps):
+  def _should_record_summaries(self, report_steps, with_accum=False):
     """Returns a boolean tensor to be used in tf.summary.record_if."""
     if report_steps is None or not self._is_master:
       return False
-    else:
-      return tf.logical_and(
-          tf.equal(self._optimizer.iterations % report_steps, 0),
+    record_summaries = tf.equal(self._optimizer.iterations % report_steps, 0)
+    if with_accum:
+      record_summaries = tf.logical_and(
+          record_summaries,
           tf.equal(self._gradient_accumulator.step, 0))
+    return record_summaries
+
+  def _compute_gradients(self, source, target, record_summaries=False):
+    """Computes the gradient of a training example."""
+    with tf.summary.record_if(record_summaries):
+      if tf.executing_eagerly():
+        with tf.GradientTape() as tape:
+          training_loss, reported_loss = self._run_model(source, target)
+        gradients = tape.gradient(training_loss, self._model.trainable_variables)
+      else:
+        training_loss, reported_loss = self._run_model(source, target)
+        gradients = self._optimizer.get_gradients(training_loss, self._model.trainable_variables)
+      _summarize_gradients(gradients, record_summaries)
+    return reported_loss, gradients
+
+  def _apply_gradients(self, gradients, scale=1):
+    """Applies the gradients."""
+    gradient_scale = scale * self.num_replicas
+    if tf.is_tensor(gradient_scale) or gradient_scale != 1:
+      gradients = [
+          self._all_reduce_sum(gradient / tf.cast(gradient_scale, gradient.dtype))
+          for gradient in gradients]
+    self._optimizer.apply_gradients(list(zip(gradients, self._model.trainable_variables)))
+
+  def _forward_and_step(self, source, target, record_summaries=False):
+    """Forwards a training example and applies the gradients, without accumulation."""
+    loss, gradients = self._compute_gradients(source, target, record_summaries=record_summaries)
+    self._apply_gradients(gradients)
+    return loss
 
   def _forward(self, source, target, record_summaries=False):
     """Forwards a training example and accumulates the gradients."""
-    with tf.summary.record_if(record_summaries):
-      training_loss, reported_loss = self._run_model(source, target)
-      gradients = self._optimizer.get_gradients(training_loss, self._model.trainable_variables)
-      self._gradient_accumulator(gradients)
-      _summarize_gradients(gradients, record_summaries)
-    return reported_loss
+    loss, gradients = self._compute_gradients(source, target, record_summaries=record_summaries)
+    self._gradient_accumulator(gradients)
+    return loss
 
   def _step(self):
     """Applies gradients and resets accumulation."""
-    gradient_scale = self._gradient_accumulator.step * self.num_replicas
-    gradients = [
-        self._all_reduce_sum(gradient / tf.cast(gradient_scale, gradient.dtype))
-        for gradient in self._gradient_accumulator.gradients]
-    self._optimizer.apply_gradients(list(zip(gradients, self._model.trainable_variables)))
+    self._apply_gradients(
+        self._gradient_accumulator.gradients,
+        scale=self._gradient_accumulator.step)
     self._gradient_accumulator.reset()
 
   def _update_words_counter(self, name, features):
@@ -260,36 +349,28 @@ class Trainer(abc.ABC):
       counter.assign(tf.constant(0, dtype=tf.int64))
     return counters
 
+  def _broadcast_variables(self):
+    """Broadcasts variables to other replicas, if required."""
+    return
+
   def _all_reduce_sum(self, value):
     """Reduces the value across all replicas."""
     return value
 
 
-class BasicTrainer(Trainer):
-  """Basic single GPU trainer."""
-
-  def _steps(self, dataset, accum_steps=1, report_steps=None):
-    if accum_steps != 1:
-      raise ValueError("BasicTrainer does not support gradient accumulation")
-    if callable(dataset):
-      dataset = dataset(tf.distribute.InputContext())
-
-    @tf.function(input_signature=dataset.element_spec)
-    def _step(source, target):
-      training_loss, reported_loss = self._run_model(source, target)
-      variables = self._model.trainable_variables
-      gradients = self._optimizer.get_gradients(training_loss, variables)
-      self._optimizer.apply_gradients(list(zip(gradients, variables)))
-      return reported_loss
-
-    for source, target in dataset:
-      yield _step(source, target)
-
-
 class HorovodTrainer(Trainer):
   """Trainer compatible with Horovod distributed training."""
 
   def __init__(self, model, optimizer, hvd, checkpoint=None):
+    """Initializes the Horovod trainer.
+
+    Args:
+      model: A :class:`opennmt.models.Model` instance to train.
+      optimizer: A ``tf.keras.optimizers.Optimizer`` instance.
+      hvd: The global Horovod object.
+      checkpoint: A :class:`opennmt.utils.checkpoint.Checkpoint` instance. If
+        not set, no checkpoints will be saved.
+    """
     super().__init__(model, optimizer, checkpoint=checkpoint, is_master=hvd.rank() == 0)
     self._hvd = hvd
 
@@ -297,53 +378,37 @@ class HorovodTrainer(Trainer):
   def num_replicas(self):
     return self._hvd.size()
 
-  def _all_reduce_sum(self, value):
-    return self._hvd.allreduce(value, op=self._hvd.Sum)
-
-  def _steps(self, dataset, accum_steps=1, report_steps=None):
+  def _finalize_dataset(self, dataset):
     if callable(dataset):
       dataset = dataset(tf.distribute.InputContext(
           num_input_pipelines=self._hvd.size(),
           input_pipeline_id=self._hvd.rank(),
           num_replicas_in_sync=self._hvd.size()))
+    return dataset
 
-    # Wrap forward and step with tf.function.
+  def _broadcast_variables(self):
+    self._hvd.broadcast_variables(self._model.variables, root_rank=0)
+    self._hvd.broadcast_variables(self._optimizer.variables(), root_rank=0)
 
-    @tf.function(input_signature=dataset.element_spec)
-    def _forward(source, target):
-      return self._forward(
-          source,
-          target,
-          record_summaries=self._should_record_summaries(report_steps))
-
-    @tf.function
-    def _step():
-      return self._step()
-
-    for i, (source, target) in enumerate(dataset):
-      loss = _forward(source, target)
-      _assert_loss_is_finite(loss)
-      if i == 0 or (i + 1) % accum_steps == 0:
-        _step()
-        if i == 0:
-          self._hvd.broadcast_variables(self._model.variables, root_rank=0)
-          self._hvd.broadcast_variables(self._optimizer.variables(), root_rank=0)
-        yield loss
+  def _all_reduce_sum(self, value):
+    return self._hvd.allreduce(value, op=self._hvd.Sum)
 
 
-class DistributionStrategyTrainer(Trainer):
-  """Trainer based on distribution strategies."""
+class MirroredStrategyTrainer(Trainer):
+  """Trainer based on ``tf.distribute.MirroredStrategy`` for local multi-GPU training."""
 
   def __init__(self, model, optimizer, checkpoint=None, devices=None):
-    """Initializes the trainer.
+    """Initializes the MirroredStrategy trainer.
 
     Args:
-      checkpoint: A :class:`opennmt.utils.checkpoint.Checkpoint` instance.
-      devices: List of device strings to use for training.
+      model: A :class:`opennmt.models.Model` instance to train.
+      optimizer: A ``tf.keras.optimizers.Optimizer`` instance.
+      checkpoint: A :class:`opennmt.utils.checkpoint.Checkpoint` instance. If
+        not set, no checkpoints will be saved.
+      devices: List of device strings to use for training. If not set, all
+        visible GPUs are used.
     """
-    super(DistributionStrategyTrainer, self).__init__(model, optimizer, checkpoint=checkpoint)
-    if not devices:
-      devices = misc.get_devices(count=1)  # Train with 1 device by default.
+    super().__init__(model, optimizer, checkpoint=checkpoint)
     self._strategy = tf.distribute.MirroredStrategy(devices=devices)
     with self._strategy.scope():
       # Create some variables under the strategy scope.
@@ -353,59 +418,31 @@ class DistributionStrategyTrainer(Trainer):
   def num_replicas(self):
     return self._strategy.num_replicas_in_sync
 
-  def _steps(self, dataset, accum_steps=1, report_steps=None):
-    for i, loss in enumerate(self._accumulate_next_gradients(dataset, report_steps=report_steps)):
-      _assert_loss_is_finite(loss)
-      if i == 0 or (i + 1) % accum_steps == 0:
-        self._apply_gradients()
-        yield loss
-
-  def _accumulate_next_gradients(self, dataset, report_steps=None):
-    """Accumulates the gradients from the next element in :obj:`dataset`."""
-
+  def _finalize_dataset(self, dataset):
     # We prefer not to use experimental_distribute_dataset here because it
     # sometimes fails to split the batches (noticed with tokens batch type).
     # We also assume for now that we are training with a single worker
     # otherwise we would need to correctly shard the input dataset.
     dataset_fn = dataset if callable(dataset) else lambda _: dataset
-    distributed_dataset = self._strategy.experimental_distribute_datasets_from_function(
-        dataset_fn)
+    return self._strategy.experimental_distribute_datasets_from_function(dataset_fn)
 
-    # Get the next element within the tf.function for more pipelining.
-    # See: https://github.com/tensorflow/tensorflow/issues/29075#issuecomment-513390242
-    iterator = iter(distributed_dataset)
-
-    @tf.function
-    def _accumulate_next():
-      record_summaries = self._should_record_summaries(report_steps)
-      per_replica_source, per_replica_target = next(iterator)
-      return self._accumulate_gradients(
-          per_replica_source,
-          per_replica_target,
-          record_summaries=record_summaries)
-
-    while True:
-      try:
-        yield _accumulate_next()
-      except tf.errors.OutOfRangeError:
-        break
-
-  def _accumulate_gradients(self,
-                            per_replica_source,
-                            per_replica_target,
-                            record_summaries=False):
-    """Accumulates the gradients (cross-replica)."""
+  def _forward_and_step(self, source, target, record_summaries=False):
     per_replica_loss = self._strategy.run(
-        self._forward,
-        args=(per_replica_source, per_replica_target),
+        super()._forward_and_step,
+        args=(source, target),
+        kwargs=dict(record_summaries=record_summaries))
+    return self._strategy.reduce(tf.distribute.ReduceOp.MEAN, per_replica_loss, None)
+
+  def _forward(self, source, target, record_summaries=False):
+    per_replica_loss = self._strategy.run(
+        super()._forward,
+        args=(source, target),
         kwargs=dict(record_summaries=record_summaries))
     # TODO: this reduction could be delayed until _step is called.
     return self._strategy.reduce(tf.distribute.ReduceOp.MEAN, per_replica_loss, None)
 
-  @tf.function
-  def _apply_gradients(self):
-    """Applies the gradients (cross-replica)."""
-    self._strategy.run(self._step)
+  def _step(self):
+    self._strategy.run(super()._step)
 
 
 def _report_training_status(step,
@@ -443,10 +480,6 @@ def _report_training_status(step,
   tf.summary.scalar("loss", loss, description="Training loss")
   tf.summary.scalar("optim/learning_rate", learning_rate, description="Learning rate")
 
-def _assert_loss_is_finite(loss):
-  if tf.math.is_nan(loss):
-    raise RuntimeError("Model diverged with loss = NaN.")
-
 def _summarize_gradients(gradients, should_record):
   # Only compute the gradients global norm when the value is actually recorded.
   if isinstance(should_record, bool) and not should_record:
@@ -499,7 +532,6 @@ class MovingAverage(object):
     Returns:
       A context manager.
     """
-    # TODO: Do we want to shadow the values on all replicas?
     previous_values = []
     for variable in self._variables:
       previous_values.append(variable.value())
@@ -507,27 +539,3 @@ class MovingAverage(object):
     yield
     for previous_value, variable in zip(previous_values, self._variables):
       variable.assign(previous_value)
-
-
-class _LossScaleOptimizer(tf.keras.mixed_precision.experimental.LossScaleOptimizer):
-  # TODO: Remove this wrapper when this fix is released:
-  # https://github.com/tensorflow/tensorflow/commit/d1dd08dd2807ac80a4508686618419826463374b
-
-  def get_unscaled_gradients(self, grads):
-    loss_scale_reciprocal = 1. / self._loss_scale()
-    return [
-        _multiply_gradient(g, loss_scale_reciprocal) if g is not None else None
-        for g in grads
-    ]
-
-
-def _multiply_gradient(gradient, scale):
-  """Multiply a (possibly sparse) gradient by the given scale factor."""
-  scale = tf.cast(scale, gradient.dtype)
-  if isinstance(gradient, tf.IndexedSlices):
-    return tf.IndexedSlices(
-        gradient.values * scale,
-        gradient.indices,
-        dense_shape=gradient.dense_shape)
-  else:
-    return gradient * scale
