@@ -1,6 +1,7 @@
 """Module defining various utilities."""
 
 from opennmt.utils.checkpoint import average_checkpoints
+from opennmt.utils.checkpoint import average_checkpoints_into_layer
 from opennmt.utils.checkpoint import is_v1_checkpoint
 
 from opennmt.utils.decoding import BeamSearch
@@ -15,6 +16,7 @@ from opennmt.utils.decoding import dynamic_decode
 from opennmt.utils.exporters import CTranslate2Exporter
 from opennmt.utils.exporters import Exporter
 from opennmt.utils.exporters import SavedModelExporter
+from opennmt.utils.exporters import register_exporter
 
 from opennmt.utils.losses import cross_entropy_loss
 from opennmt.utils.losses import cross_entropy_sequence_loss
